from django.apps import AppConfig


class DjangoAdvancedPdfConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'django_advanced_pdf'
    verbose_name = 'Django advanced pdfs'
